#include <stdio.h>
#include <stdarg.h>
#include "tree.h"
#include "symbol_table.h"
#include "prover.h"
#include "newdict.h"


extern void create_clauses (int set, tnode *s);
extern literalslist *simplify_disjunction(literal_list *l);
extern int size_clause (literalslist *l);
extern int max_lit_clause (literalslist *l);
extern int classify_clause (literalslist *l);
extern void insert_clause_set (tclause **clause, int type, int set);

extern int is_literal (tnode *);
extern tnode *create_tnode(int type, int id, int mdepth, tnode *left, tnode *right, formulalist *list);
extern void copy_node(tnode *t1, tnode *t2);
extern tnode *copy_tree(tnode *s);
extern int same_tree(tnode *t1, tnode *t2); 
extern tnode *free_tree (tnode *s);


extern tnode *tree_hash(tnode *node);
extern unsigned int hash_tree (tnode *t);
//extern unsigned int hash_list (formulalist *l);

extern agent_node *find_agent (int id);   
extern void create_formula_dict(tnode *formula, int underbox);
extern ren new_renaming (tnode *formula, int mlevel);

extern int numprops;

extern int ires;
extern int ml_prover;
extern int numclause;
extern int global2local;
extern int mdepth;
extern int mlbound;

justification *create_justification(int r, clauseslist *p, literalslist *l) {
  justification *just = malloc(sizeof(justification));
  just->rule = r;
  just->parents = p;
  just->literals = l;
  return just;
}


formulalist *create_disjunction(int n, tnode *t,...) {
  va_list va;
  int i;
  
  va_start(va,t);
  formulalist *list = NULL;
  for (i = 0; i < n; i++) {
    formulalist *new = malloc(sizeof(formulalist));
    new->formula = t;
    t = va_arg(va,tnode*);
    new->next = list;
    list = new;
  }
  va_end(va);
  
  return list;
}


tnode *negate_literal(tnode *s) {
  if (s->type == NEGATION) return s->left;
  else return create_tnode(NEGATION,NEGATION,0,s,NULL,NULL);
}

tnode *create_tnode_lit(int id){
  tnode *new = create_tnode(PROP,id,0,NULL,NULL,NULL);
  if (id < 0) {
    new->id = -new->id;
    tnode *neg = create_tnode(NEGATION,NEGATION,0,new,NULL,NULL);
    new = neg;
  }
  return new;
}

void create_clause_snf_local(int set, int type, int subtype, int modal_level, int agent, int numdiamond, tnode *left, formulalist  *right, justification *just) {

  tclause *clause = malloc(sizeof(tclause));
  clause->number = numclause++;
  clause->type = type;
  
  clause->subtype = subtype;
  clause->numdiamond = numdiamond;
  
  if (!ml_prover)
    clause->modal_level = 1;
  else
    clause->modal_level = modal_level;
  clause->agent = agent;

  if (type == INITIAL) clause->left = CSTART;
  else if (type == LITERAL) clause->left = CTRUE;
  else if (left->type == NEGATION) clause->left = -left->left->id;
  else clause->left = left->id;

  literal_list *dis = NULL;
  if (left->type != CONSTANT && (type == LITERAL || type == INITIAL)) {
    dis = malloc(sizeof(literal_list));
    dis->literal = negate_literal(left);
    dis->next = NULL;
  }
 
  formulalist *aux = right;
  while (aux != NULL) {
    literal_list *newdis = malloc(sizeof(literal_list));
    newdis->literal = aux->formula;
    newdis->next = dis;
    dis = newdis;
    aux = aux->next;
  }
    
  clause->right = simplify_disjunction(dis);
  clause->just = just;
  clause->deleted = NULL;
  clause->propagated = 0;
  clause->size = size_clause(clause->right);
  clause->max_literal = max_lit_clause(clause->right);
  clause->min_literal = clause->right->literal;
  clause->class = classify_clause(clause->right);
  insert_clause_set(&clause,clause->type,set);
}


/* The input formula is an implication */
int get_snf_local (int set, int type, int subtype, int mlevel, tnode *s, int recur) {

  if (s == NULL)
    return 0;
  if (s->type != IMPLICATION) {
    printf("\n SNF is taking argument of wrong type: %d \n", s->type);
  }
  else {
    switch (s->right->type) {
    case PROP:
    case CONSTANT:
    case NEGATION: // this must be a literal
	{
	  //tnode *left = copy_tree(s->left);
	  tnode *left = s->left;
	  formulalist *aux = malloc(sizeof(formulalist));
	  //aux->formula = copy_tree(s->right);
	  aux->formula = s->right;
	  aux->next = NULL;

	  justification *just = create_justification(recur,NULL,NULL);
	  create_clause_snf_local(set,LITERAL,subtype,mlevel,PROP,1,left,aux,just);
	  free(aux);
	}
      break;
    case CONJUNCTION:
      {
	formulalist *list = s->right->list;
	while (list != NULL) {
	  //	  tnode *left1 = copy_tree(s->left);
	  int flag = 0;
	  tnode *right1 = NULL;
	  if (!(is_literal(list->formula))) {
	    ren def = new_renaming(list->formula,mlevel);
	    flag = 1;
	    right1 = create_tnode(PROP,def.renamed_by,s->mdepth,NULL,NULL,NULL);
	    if (!def.added_to_level) {
	      //	      tnode *left2 = copy_tree(right1);
	      tnode *left2 = right1;
	      //	      tnode *right2 = copy_tree(list->formula);
	      tnode *right2 = list->formula;
	      tnode *newimp2 = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,left2,right2,NULL);
	      get_snf_local(set,type,subtype,mlevel,newimp2,recur);
	      free(newimp2);
	    }
	  }
	  //	  else right1 = copy_tree(list->formula);
	  else right1 = list->formula;
	  tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,s->left,right1,NULL);
	  get_snf_local(set,type,subtype,mlevel,newimp,recur);
	  list = list->next;
	  free(newimp);
	  if (flag) free_tree(right1);
	}
      }
      break;
    case DISJUNCTION:
      {
	formulalist *right = s->right->list;
	formulalist *newlist = NULL;
	while (right != NULL) {
	  if (!(is_literal(right->formula))) {
	    ren def = new_renaming(right->formula,mlevel);
	    tnode *new_lit_1 = create_tnode_lit(def.renamed_by);
	    if (!def.added_to_level) {
	      tnode *new_lit_2 = create_tnode_lit(def.renamed_by);
	      //	      tnode *right_2 = copy_tree(right->formula);
	      tnode *right_2 = right->formula;
	      tnode *new_imp2 = create_tnode(IMPLICATION,IMPLICATION,s->right->mdepth,new_lit_2,right_2,NULL);
	      get_snf_local(set,type,subtype,mlevel,new_imp2,recur);
	      free(new_imp2); free_tree(new_lit_2);
	    }
	    formulalist *aux = malloc(sizeof(formulalist));
	    aux->formula = new_lit_1;
	    aux->next = newlist;
	    newlist = aux;
	  }
	  else {
	    formulalist *aux = malloc(sizeof(formulalist));
	    //	    aux->formula = copy_tree(right->formula);
	    aux->formula = copy_tree(right->formula); //I'm keeping the copy of the literal, so that I know that I can delete the whole newlist afterwards;
	    aux->next = newlist;
	    newlist = aux;
	  }
	  right = right->next;
	}

	justification *just = create_justification(recur,NULL,NULL);
	//	tnode *left = copy_tree(s->left);
	tnode *left = s->left;
	create_clause_snf_local(set,LITERAL,subtype,mlevel,PROP,1,left,newlist,just);
	formulalist *aux = newlist;
	while (aux != NULL) {
	  formulalist *auxl = aux->next;
	  free_tree(aux->formula);
	  free(aux);
	  aux = auxl;
	}
      }
      break;
    case DIAMOND:
    case BOX:
      {
	int flag = 0;
	int clause_type = MPOSITIVE;
	if (s->right->type == DIAMOND) clause_type = MNEGATIVE;
	int clause_subtype = s->right->numdiamond;

	tnode *new_right = NULL;
	agent_node *a = find_agent(s->right->id);
	
	if (!(is_literal(s->right->left))) {
	  ren def = new_renaming(s->right->left,mlevel+1);
	  flag = 1;
	  new_right = create_tnode_lit(def.renamed_by);
	  if (!def.added_to_level) {
	    tnode *new_lit_2 = create_tnode_lit(def.renamed_by);
	    tnode *new_imp = create_tnode(IMPLICATION,IMPLICATION,s->right->mdepth,new_lit_2,s->right->left,NULL);
	    if (ml_prover == ON) 
	      get_snf_local(set,clause_type,clause_subtype,mlevel+1,new_imp,recur);
	    else get_snf_local(set,clause_type,clause_subtype,1,new_imp,recur);
	    free_tree(new_lit_2);
	  }
	}
	else new_right = s->right->left;
	
	justification *just = create_justification(recur,NULL,NULL);
	formulalist *list = malloc(sizeof(formulalist));
	list->formula = new_right;
	list->next = NULL;

	create_clause_snf_local(set,clause_type,clause_subtype,mlevel,s->right->id,s->left->numdiamond,s->left,list,just);
	if (flag) free_tree(new_right);
	free(list);

	if (clause_type == MPOSITIVE && ((a->axioms & T) != 0) && (recur != T) ) {
	  if (!same_tree(s->left,s->right->left)) {
	    if (is_literal(s->right->left)) {
	      tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,s->left,s->right->left,NULL);
	      get_snf_local(set,LITERAL,clause_subtype,mlevel,newimp,T);
	      free(newimp);
	    }
	    else {
	      ren def = new_renaming(s->right->left,mlevel+1);
	      tnode *right1 = create_tnode(PROP,def.renamed_by,s->mdepth,NULL,NULL,NULL);
	      tnode *imp1 = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,s->left,right1,NULL);
	      get_snf_local(set,LITERAL,clause_subtype,mlevel,imp1,T);
	      def = new_renaming(s->right->left,mlevel);
	      if (!def.added_to_level) {
		tnode *imp2 = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,right1,s->right->left,NULL);
		get_snf_local(set,LITERAL,clause_subtype,mlevel,imp2,recur);
		free(imp2);
	      }
	      free_tree(right1); free(imp1);
	    }
	  }
	}
	if (clause_type == MPOSITIVE && ((a->axioms & SER) != 0) && (recur != SER)) {
	  tnode *dia = create_tnode(DIAMOND,s->right->id,s->mdepth,s->right->left,NULL,NULL);
	  // Why do I need the following?
	  dia->value_number = hash_tree(dia);
	  create_formula_dict(dia,0); //underbox doesn't matter here, so it is set to zero.
	  tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,s->left,dia,NULL);
	  get_snf_local(set,MNEGATIVE,clause_subtype,mlevel,newimp,SER);
	  free(newimp);
	}

	if (clause_type == MPOSITIVE && ((a->axioms & FOUR1) != 0) && (recur != FOUR1) && (recur != SYM) ) {
	  int levels = 0;
	  ren def = new_renaming(s->right,mlevel);

	  if (def.renamed_by != s->left->id) { // t -> []q, but t2 renames []q, so we add (~t | t2)
	    tnode *left = create_tnode(PROP,def.renamed_by,mlevel,NULL,NULL,NULL);
	    tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,mlevel,left,s->left,NULL);
	    get_snf_local(set,LITERAL,clause_subtype,mlevel,newimp,FOUR1);
	    free(newimp);
	    }

	  int loop_bound;
	  if (mlbound != 0) loop_bound = mlbound; else loop_bound = 2 + a->maxleveldia + ((a->numdifdia)*(a->numdifbox));
	  // The following was set experimentally for the CADE-28 paper, considering the benchmarks we had
	  // if (mlbound != 0) loop_bound = mlbound; else loop_bound = mdepth + 2;
	  for (levels = mlevel; levels <= loop_bound; levels++) { //t->[]q, t2 renames []q
	    
	    // add t2 => []q to levels
	    if (levels > mlevel) { // this is added by K in the current level
	      tnode *left1 = create_tnode(PROP,def.renamed_by,levels,NULL,NULL,NULL);
	      tnode *newimp1 = create_tnode(IMPLICATION,IMPLICATION,levels,left1,s->right,NULL);
	      get_snf_local(set,MPOSITIVE,clause_subtype,levels,newimp1,FOUR1);
	      free_tree(left1); free(newimp1); 
	    }
	    
	      // add t2 =>[]t2 to levels
	      tnode *left2 = create_tnode(PROP,def.renamed_by,levels,NULL,NULL,NULL); //add t2 ->[]t2 to levels
	      tnode *right2 = create_tnode(PROP,def.renamed_by,levels,NULL,NULL,NULL);
	      tnode *box = create_tnode(BOX,s->right->id,levels,right2,NULL,NULL);
	      tnode *newimp2 = create_tnode(IMPLICATION,IMPLICATION,levels,left2,box,NULL);
	      get_snf_local(set,MPOSITIVE,clause_subtype,levels,newimp2,FOUR1);
	      free_tree(newimp2);
	   
	      if (levels > mlevel && !(is_literal(s->right->left))) { // q is not a literal, then definition needs to be added to levels+1, but not in the case of the current level, as this is added by K.
		ren newdef = new_renaming(s->right->left,levels+1);
		if (newdef.renamed_by == 0) {
		  if (s->right->left->type == NEGATION) { newdef.renamed_by = s->right->left->left->id;}
		  else newdef.renamed_by = s->right->left->id;
		}
		if (!(is_literal(s->right->left))) {
		  tnode *left4 = create_tnode(PROP,newdef.renamed_by,levels+1,NULL,NULL,NULL);
		  tnode *newimpred = create_tnode(IMPLICATION,IMPLICATION,levels+1,left4,s->right->left,NULL);
		  get_snf_local(set,LITERAL,clause_subtype,levels+1,newimpred,FOUR1);
		  free_tree(left4); free(newimpred);
		}
	      }
	  }
	}

	if (clause_type == MPOSITIVE && ((a->axioms & FIVE) != 0) && (recur != FIVE)) {
	  int levels = 0;
	  ren def = new_renaming(s->right,mlevel);
	  if (def.renamed_by != s->left->id) {
	    //	    printf("\n is entering here");
	    tnode *left = create_tnode(PROP,def.renamed_by,mlevel,NULL,NULL,NULL);
	    tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,mlevel,left,s->left,NULL);
	    get_snf_local(set,LITERAL,clause_subtype,mlevel,newimp,FIVE);
	    free(newimp);
	  }
	  int loop_bound;
	  if (mlbound != 0) loop_bound = mlbound; else loop_bound = 1 + a->maxleveldia + a->numdifdia;
	  // The following was set experimentally for the CADE-28 paper, considering the benchmarks we had
	  // if (mlbound != 0) loop_bound = mlbound; else loop_bound = mdepth + 2;
	  for (levels = 1; levels <= loop_bound; levels++) {
	    if (levels != mlevel) { // definition is added to the current level by K; Note that this is being added to all levels
	      tnode *left1 = create_tnode(PROP,def.renamed_by,levels,NULL,NULL,NULL);
	      tnode *newimp1 = create_tnode(IMPLICATION,IMPLICATION,levels,left1,s->right,NULL);
	      get_snf_local(set,MPOSITIVE,clause_subtype,levels,newimp1,FIVE);
	      free_tree(left1); free(newimp1);
	    }
	    
	    tnode *right2 = create_tnode(PROP,def.renamed_by,levels,NULL,NULL,NULL);
	    tnode *dia = create_tnode(DIAMOND,s->right->id,levels,right2,NULL,NULL);
	    dia->value_number = hash_tree(dia);
	    create_formula_dict(dia,0); //underbox doesn't matter here, so it is set to zero.
	    ren newdef = new_renaming(dia,levels);
	    tnode *left2 = create_tnode(PROP,newdef.renamed_by,levels,NULL,NULL,NULL);
	    tnode *newimp2 = create_tnode(IMPLICATION,IMPLICATION,levels,left2,dia,NULL);
	    get_snf_local(set,MNEGATIVE,clause_subtype,levels,newimp2,FIVE);
	    free_tree(left2); free(newimp2);
	    // cannot free the tree, because the diamond is in the dictionary

	    tnode *right3 = create_tnode(PROP,def.renamed_by,levels,NULL,NULL,NULL);
	    tnode *neg1 = create_tnode(NEGATION,NEGATION,levels,right3,NULL,NULL);
	    tnode *box = create_tnode(BOX,s->right->id,levels,neg1,NULL,NULL);
	    tnode *left3 = create_tnode(PROP,newdef.renamed_by,levels,NULL,NULL,NULL);
	    tnode *neg2 = create_tnode(NEGATION,NEGATION,levels,left3,NULL,NULL);
	    tnode *newimp3 = create_tnode(IMPLICATION,IMPLICATION,levels,neg2,box,NULL);
	    get_snf_local(set,MPOSITIVE,clause_subtype,levels,newimp3,FIVE);
	    free_tree(newimp3);

	    tnode *left4 = create_tnode(PROP,newdef.renamed_by,levels,NULL,NULL,NULL);
	    tnode *right4 = create_tnode(PROP,newdef.renamed_by,levels,NULL,NULL,NULL);
	    tnode *box2 = create_tnode(BOX,s->right->id,levels,right4,NULL,NULL);
	    tnode *newimp4 = create_tnode(IMPLICATION,IMPLICATION,levels,left4,box2,NULL);
	    get_snf_local(set,MPOSITIVE,clause_subtype,levels,newimp4,FIVE);
	    free_tree(newimp4);

	    tnode *left5 = create_tnode(PROP,newdef.renamed_by,levels,NULL,NULL,NULL);
	    tnode *right5 = create_tnode(PROP,def.renamed_by,levels,NULL,NULL,NULL);
	    tnode *newimp5 = create_tnode(IMPLICATION,IMPLICATION,levels,left5,right5,NULL);
	    get_snf_local(set,LITERAL,clause_subtype,levels,newimp5,FIVE);
	    free_tree(newimp5);
	    
	    if (levels != mlevel && !(is_literal(s->right->left))) {
	      ren newdef2 = new_renaming(s->right->left,levels+1);
	      if (newdef2.renamed_by == 0) {
		if (s->right->left->type == NEGATION)  newdef2.renamed_by = s->right->left->left->id;
		else newdef2.renamed_by = s->right->left->id;
	      }
	      if (!(is_literal(s->right->left))) {
		tnode *left6 = create_tnode(PROP,newdef2.renamed_by,levels,NULL,NULL,NULL);
		tnode *newimpred = create_tnode(IMPLICATION,IMPLICATION,levels+1,left6,s->right->left,NULL);
		get_snf_local(set,LITERAL,clause_subtype,levels+1,newimpred,FIVE);
		free_tree(left6);free(newimpred);
	      }
	    }
	  }
	}

	if (mlevel != 1 && clause_type == MPOSITIVE && ((a->axioms & SYM) != 0) && (recur != SYM)) {
	  /* t_([a]~t) => [a]~t : in the predecessor */
	  tnode *neg = create_tnode(NEGATION,NEGATION,s->mdepth,s->left,NULL,NULL);
	  tnode *box = create_tnode(BOX,s->right->id,s->mdepth,neg,NULL,NULL);
	  box->value_number = hash_tree(box);
	  create_formula_dict(box,0); //underbox doesn't matter here, so it is set to zero.
	  ren def = new_renaming(box,mlevel);
	  tnode *propdef = create_tnode(PROP,def.renamed_by,s->mdepth,NULL,NULL,NULL);
	  tnode *newimpdef = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,propdef,box,NULL);
	  get_snf_local(set,MPOSITIVE,clause_subtype,mlevel-1,newimpdef,SYM);
	  free(newimpdef);

	  /* t_phi \/ t_([a]~t) : in the predecessor */
	  tnode *left = create_tnode(CONSTANT,CTRUE,s->mdepth,NULL,NULL,NULL);
	  tnode *right2 = propdef;
	  formulalist *newlist = create_disjunction(2,s->right->left,right2);
	  tnode *disj = create_tnode(DISJUNCTION,DISJUNCTION,s->mdepth,NULL,NULL,newlist);
	  tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,left,disj,NULL);
	  get_snf_local(set,LITERAL,clause_subtype,mlevel-1,newimp,SYM);
	  free(newimp);

	  /* t_phi => phi : in the predecessor */
	  if (!(is_literal(s->right->left))) {
	    def = new_renaming(s->right->left,mlevel-1);
	    if (def.renamed_by == 0) {
	      if (s->right->left->type == NEGATION)  def.renamed_by = s->right->left->left->id;
	      else def.renamed_by = s->right->left->id;
	    }
	    if (!(is_literal(s->right->left))) {
	      tnode *left3 = create_tnode(PROP,def.renamed_by,s->mdepth,NULL,NULL,NULL);
	      tnode *newimpred = create_tnode(IMPLICATION,IMPLICATION,s->mdepth,left3,s->right->left,NULL);
	      get_snf_local(set,LITERAL,clause_subtype,mlevel-1,newimpred,SYM);
	      free(newimpred);
	    }
	  }
	}
      }
      break;
    }
  }
  return 0;
}



int anchor_snf_local (tnode *root, int set, int mlevel) {
  if (root == NULL) return 0;
  switch (root->type) {
  case SETC: {
    create_clauses(SETC,root->left);
    if (root->right != NULL) 
      anchor_snf_local(root->right, set, mlevel);
  } break;
  case SETF: {
    anchor_snf_local(root->left, SETF, mlevel);
    if (root->right != NULL) 
      anchor_snf_local(root->right, set, mlevel);
  } break;
  default: {
      ren def = new_renaming(root,1);
      tnode *new1 = create_tnode(PROP,def.renamed_by,0,NULL,NULL,NULL);
      tnode *new2 = create_tnode(PROP,def.renamed_by,0,NULL,NULL,NULL);

      formulalist *dis = create_disjunction(1,new1);
      justification *just = create_justification(SNF,NULL,NULL);

      int type = LITERAL;
      tnode *left = create_tnode(CONSTANT,CTRUE,0,NULL,NULL,NULL);
      if (ires) {
	type = INITIAL;
	left->id = CSTART;
      }
      
      create_clause_snf_local(set,type,BOX,mlevel,PROP,1,left,dis,just);
      tnode *newimp = create_tnode(IMPLICATION,IMPLICATION,root->mdepth,new2,root,NULL);
      get_snf_local(set,LITERAL,BOX,mlevel,newimp,SNF);
      
      free(newimp); free(new1); free(new2); free(dis);
    } break;
  }
  return 1;
}
